using LinearAlgebra: dot, pinv, norm, cond, eigmax, eigmin, svdvals

import ForwardDiff
using ForwardDiff: jacobian, derivative, gradient
using Krylov
using LinearOperators: LinearOperator
using StaticArrays
using StaticKernels: Kernel, ExtensionReplicate, ExtensionNothing


struct TVNewtonAlgorithm{P} <: Algorithm{P}
    problem::P

    function TVNewtonAlgorithm(problem)
        return new{typeof(problem)}(problem)
    end
end

struct TVNewtonState{A,Tx,Tu,Tn,K1,K2}
    algorithm::A

    "combined primal/dual variable"
    x::Tx
    "primal view on `x`"
    u::Tu
    "dual view on `x`"
    n::Tn

    kdiv::K1
    kgrad::K2
end

function init(alg::TVNewtonAlgorithm{<:DualTVL1ROFOpProblem})
    g = alg.problem.g
    λ = alg.problem.λ
    sz = size(g)
    d = ndims(g)
    k = eltype(g) <: StaticArray ? length(eltype(g)) : 1

    len = k * prod(sz), d * k * prod(sz)
    x = Vector{eltype(eltype(g))}(undef, sum(len))
    u, n = interpret(x, sz)

    #x .= rand(size(x))
    u .= g
    n .= Ref(zero(eltype(n)))
    #n .= rand(eltype(n), size(n))

    kgrad = Kernel{ntuple(_->0:1, d)}(gradient)
    kdiv = Kernel{ntuple(_->-1:1, d)}(mydivergence)

    return TVNewtonState(alg, x, u, n, kdiv, kgrad)
end

function interpret(x, sz)
    d = length(sz)
    n = length(x)
    k = n ÷ prod(sz) ÷ (d + 1)

    T = eltype(x)
    Tu = k == 1 ? T : SVector{k,T}
    Tn = SVector{d,Tu}

    len = k * prod(sz), d * k * prod(sz)

    uview = extend(reshape(reinterpret(Tu, @view(x[begin:len[1]])), sz), ExtensionReplicate())
    nview = extend(reshape(reinterpret(Tn, @view(x[len[1]+1:end])), sz), ExtensionNothing())

    return uview, nview
end


function step!(st::TVNewtonState)
    alg = st.algorithm
    prob = alg.problem

    g, λ, B = prob.g, prob.λ, prob.B
    x = st.x

    function f(x)
        u, n = interpret(x, size(prob.g))
        ugrad = map(st.kgrad, u)
        ndiv = map(st.kdiv, n)

        fu = u .- reshape(B * vec(λ .* ndiv .+ prob.g), size(u))
        #fu = reshape(B \ vec(u), size(u)) .- prob.g .- λ .* ndiv
        #fn = sqrt.(norm2.(ugrad) .+ floatmin()) .* ugrad .- norm2.(ugrad) .* n
        fn = ugrad .- sqrt.(norm2.(ugrad) .+ floatmin()) .* n

        T = eltype(x)
        return vcat(reinterpret(T, vec(fu)), reinterpret(T, vec(fn)))
    end

    fx = f(x)
    A = jacobian(f, x)
    #any(isnan, A) && error("jacobian contains NaN")

    Amap = LinearOperator(eltype(x), length(fx), length(x), false, false,
        y -> jvp(f, x, y), y -> vjp(f, x, y), y -> vjp(f, x, y))

    #println("cond(A) = $(cond(A))")
    #println("svdvals(A) = $(svdvals(A))")
    #dir = -dqgmres(Amap, fx)[1]
    dir = -cgs(Amap, fx)[1]
    #dir = -cg_lanczos(Amap, fx)[1]
    #dir = -lsqr(Amap, fx)
    #dir = -(A \ fx)
    #dir = -(pinv(A) * fx)
    τ = 1.
    y = x + τ * dir
    project!(y, size(g))
    #while norm(f(y)) >= norm(fx)
    #    τ /= 2
    #    y = x + τ * dir
    #    project!(y, size(g))
    #    τ == 0 && break
    #end
    println("residual norm : $(norm(f(y))/length(fx)) with τ = $τ")
    st.x .= y

    return st
end

fetch(ctx::TVNewtonState) = ctx.algorithm.problem.λ .* ctx.n

@generated function mydivergence(w::StaticKernels.Window{SVector{N,T},N}) where {N,T}
    i0 = ntuple(_->0, N)
    i1(k) = ntuple(i->Int(k==i), N)

    wi = (:((isnothing(w[$(i1(k)...)]) ? zero(T) : w[$(i0...)][$k]) -
            (isnothing(w[$((.-i1(k))...)]) ? zero(T) : w[$((.-i1(k))...)][$k])) for k in 1:N)
    return quote
        Base.@_inline_meta
        return @inbounds +($(wi...))
    end
end

relu(x) = x < 0 ? 0 : x
norm2(x) = dot(x,x)

function project!(x, sz)
    _, n = interpret(x, sz)
    fp(y) = norm(y) > 1 ? y/norm(y) : y
    n .= fp.(n)
end

jvp(f, x, v) = ForwardDiff.derivative(ε -> f(x .+ ε .* v), 0.)
vjp(f, x, v) = ForwardDiff.gradient(φ -> dot(f(φ), v), x)

function newton(f, x, g)
end

using Revise
Revise.track(@__FILE__)
